@@ -31,84 +31,99 @@ class HTTP:
     Pointer, go to http://tools.ietf.org/html/draft-pbryan-zyp-json-pointer-00.
     """
 
+    ROBOT_LIBRARY_VERSION = "0.3.0"
+
+    class Context(object):
+        def __init__(self, http, host=None):
+            # daddy
+            self._http = http
+
+            # the livetest app
+            if host:
+                self.app = livetest.TestApp(host)
+            else:
+                self.app = None
+
+            # the last request
+            self.response = None
+
+            # requirements for the next request
+            # None -> no requirements
+            # True -> request should succeed
+            # False -> request should not succeed
+            # string -> response status code should startwith(string)
+            self.next_request_should = None
+
+            # setup new http context
+            self.post_process_request(None)
+
+        def pre_process_request(self):
+            if len(self.request_headers.items()) > 0:
+                logger.debug("Request headers:")
+                for name, value in self.request_headers.items():
+                    logger.debug("%s: %s" % (name, value))
+            else:
+                logger.debug("No request headers set")
+
+            if self.request_body is None:
+                logger.debug("No request body set")
+            else:
+                logger.debug("Request body:")
+                logger.debug(self.request_body)
+
+        def post_process_request(self, response):
+            self.response = response
+
+            if response != None:
+                self._http.log_response_status('DEBUG')
+                self._http.log_response_headers('DEBUG')
+                self._http.log_response_body('DEBUG')
+
+            next_request_should = self.next_request_should
+
+            # prepare next request context, even if one of the following assertions
+            # fail
+            self.next_request_should = True
+            self.request_headers = {}
+            self.request_body = None
+
+            # check flag set by "Next Request Should Succeed"
+            if next_request_should == True:
+                assert int(self.response.status[0:3]) < 400, \
+                   'Request should have succeeded, but was "%s".' % \
+                   self.response.status
+
+            # check flag set by "Next Request Should Not Succeed"
+            elif next_request_should == False:
+                assert int(self.response.status[0:3]) >= 400, \
+                   'Request should not have succeeded, but was "%s".' % \
+                   self.response.status
+
+            elif next_request_should:
+                self._http.response_status_code_should_equal(next_request_should)
+
     ROBOT_LIBRARY_SCOPE = 'TEST SUITE'
 
     # internal
 
     def __init__(self):
+        self._contexts = [HTTP.Context(self)]
 
-        # the livetest http context
-        self._app = None
-
-        # the last request
-        self._response = None
-
-        # requirements for the next request
-        # None -> no requirements
-        # True -> request should succeed
-        # False -> request should not succeed
-        # string -> response status code should startwith(string)
-        self._next_request_should = None
-
-        # setup new http context
-        self._post_process_request()
-
-    def _pre_process_request(self):
-
-        if len(self._request_headers.items()) > 0:
-            logger.debug("Request headers:")
-            for name, value in self._request_headers.items():
-                logger.debug("%s: %s" % (name, value))
-        else:
-            logger.debug("No request headers set")
-
-        if self._request_body is None:
-            logger.debug("No request body set")
-        else:
-            logger.debug("Request body:")
-            logger.debug(self._request_body)
-
-    def _post_process_request(self):
-
-        if self._response != None:
-            self.log_response_status('DEBUG')
-            self.log_response_headers('DEBUG')
-            self.log_response_body('DEBUG')
-
-        next_request_should = self._next_request_should
-
-        # prepare next request context, even if one of the following assertions
-        # fail
-        self._next_request_should = True
-        self._request_headers = {}
-        self._request_body = None
-
-        # check flag set by "Next Request Should Succeed"
-        if next_request_should == True:
-            assert int(self.response.status[0:3]) < 400, \
-               'Request should have succeeded, but was "%s".' % \
-               self.response.status
-
-        # check flag set by "Next Request Should Not Succeed"
-        elif next_request_should == False:
-            assert int(self.response.status[0:3]) >= 400, \
-               'Request should not have succeeded, but was "%s".' % \
-               self.response.status
-
-        elif next_request_should:
-            self.response_status_code_should_equal(next_request_should)
+    @property
+    def context(self):
+        return self._contexts[-1]
 
     @property
     def app(self):
-        if not self._app:
+        if not self.context.app:
             raise Exception('Not connected to any HTTP Host. Use "Set HTTP Host" keyword first.')
-        return self._app
+        return self.context.app
 
     @property
     def response(self):
-        if not self._response:
+        if not self.context.response:
             raise Exception('No request available, use e.g. GET to create one.')
-        return self._response
+        return self.context.response
 
     def _path_from_url_or_path(self, url_or_path):
 
@@ -126,14 +141,32 @@ class HTTP:
     # setup
 
     def set_http_host(self, host):
+        """
+        (deprecated) Sets the HTTP host to use for future requests. You must call this
+        before issuing any HTTP requests.
+
+        `host` is the name of the host, optionally with port (e.g. 'google.com' or 'localhost:5984')
+        """
+        self.create_http_context(host)
+
+    def create_http_context(self, host=None):
         """
         Sets the HTTP host to use for future requests. You must call this
         before issuing any HTTP requests.
 
         `host` is the name of the host, optionally with port (e.g. 'google.com' or 'localhost:5984')
         """
+        if host == None:
+            host = self.context.app.host
         logger.info("Host for next HTTP request set to '%s'" % host)
-        self._app = livetest.TestApp(host)
+        self._contexts.append(HTTP.Context(self, host))
+
+    def restore_http_context(self):
+        """
+        Restores HttpLibrary's state to the point before the last "Create Http
+        Context" call.
+        """
+        self._contexts.pop()
 
     # request
 
@@ -146,11 +179,12 @@ class HTTP:
         """
         path = self._path_from_url_or_path(url)
 
-        self._pre_process_request()
+        self.context.pre_process_request()
         logger.debug("Performing %s request on http://%s%s" % (verb, self.app.host, url,))
-        self._response = self._app.request(path, {}, self._request_headers,
-                method=verb.upper(),)
-        self._post_process_request()
+        self.context.post_process_request(
+            self.context.app.request(path, {}, self.context.request_headers,
+            method=verb.upper(),)
+        )
 
     def HEAD(self, url):
         """
@@ -159,10 +193,11 @@ class HTTP:
         `url` is the URL relative to the server root, e.g. '/_utils/config.html'
         """
         path = self._path_from_url_or_path(url)
-        self._pre_process_request()
+        self.context.pre_process_request()
         logger.debug("Performing HEAD request on http://%s%s" % (self.app.host, url,))
-        self._response = self.app.head(path, {}, self._request_headers)
-        self._post_process_request()
+        self.context.post_process_request(
+          self.app.head(path, {}, self.context.request_headers)
+        )
 
     def GET(self, url):
         """
@@ -171,10 +206,11 @@ class HTTP:
         `url` is the URL relative to the server root, e.g. '/_utils/config.html'
         """
         path = self._path_from_url_or_path(url)
-        self._pre_process_request()
+        self.context.pre_process_request()
         logger.debug("Performing GET request on http://%s%s" % (self.app.host, url))
-        self._response = self.app.get(path, {}, self._request_headers)
-        self._post_process_request()
+        self.context.post_process_request(
+          self.app.get(path, {}, self.context.request_headers)
+        )
 
     def POST(self, url):
         """
@@ -184,12 +220,13 @@ class HTTP:
         """
         path = self._path_from_url_or_path(url)
         kwargs = {}
-        if 'Content-Type' in self._request_headers:
-            kwargs['content_type'] = self._request_headers['Content-Type']
+        if 'Content-Type' in self.context.request_headers:
+            kwargs['content_type'] = self.context.request_headers['Content-Type']
         logger.debug("Performing POST request on http://%s%s" % (self.app.host, url))
-        self._pre_process_request()
-        self._response = self.app.post(path, self._request_body or {}, self._request_headers, **kwargs)
-        self._post_process_request()
+        self.context.pre_process_request()
+        self.context.post_process_request(
+          self.app.post(path, self.context.request_body or {}, self.context.request_headers, **kwargs)
+        )
 
     def PUT(self, url):
         """
@@ -199,12 +236,13 @@ class HTTP:
         """
         path = self._path_from_url_or_path(url)
         kwargs = {}
-        if 'Content-Type' in self._request_headers:
-            kwargs['content_type'] = self._request_headers['Content-Type']
-        self._pre_process_request()
+        if 'Content-Type' in self.context.request_headers:
+            kwargs['content_type'] = self.context.request_headers['Content-Type']
+        self.context.pre_process_request()
         logger.debug("Performing PUT request on http://%s%s" % (self.app.host, url))
-        self._response = self.app.put(path, self._request_body or {}, self._request_headers, **kwargs)
-        self._post_process_request()
+        self.context.post_process_request(
+          self.app.put(path, self.context.request_body or {}, self.context.request_headers, **kwargs)
+        )
 
     def DELETE(self, url):
         """
@@ -213,10 +251,11 @@ class HTTP:
         `url` is the URL relative to the server root, e.g. '/_utils/config.html'
         """
         path = self._path_from_url_or_path(url)
-        self._pre_process_request()
+        self.context.pre_process_request()
         logger.debug("Performing DELETE request on %s" % url)
-        self._response = self.app.delete(path, {}, self._request_headers)
-        self._post_process_request()
+        self.context.post_process_request(
+          self.app.delete(path, {}, self.context.request_headers)
+        )
 
     def follow_response(self):
         """
@@ -230,34 +269,34 @@ class HTTP:
 
         logger.debug("Following response, last response's Location header was %s" % location)
 
-        self._response = self.response.follow()
+        self.context.response = self.response.follow()
 
 
-    def next_request_may_not_succeed(self, status_code=None):
+    def next_request_may_not_succeed(self):
         """
         Don't fail the next request if it's status code is >= 400
         """
-        self._next_request_should = None
+        self.context.next_request_should = None
 
-    def next_request_should_succeed(self, status_code=None):
+    def next_request_should_succeed(self):
         """
         Fails the next request if it's status code is >= 400. This is the
         standard behaviour (only use this keyword if you specified `Next
         Request Should Not Succeed` earlier.
         """
-        self._next_request_should = True
+        self.context.next_request_should = True
 
     def next_request_should_not_succeed(self):
         """
         Fails the next request if it's status code is < 400
         """
-        self._next_request_should = False
+        self.context.next_request_should = False
 
     def next_request_should_have_status_code(self, status_code=None):
         """
         Fails the next request if it's status code is different from `status_code`.
         """
-        self._next_request_should = status_code
+        self.context.next_request_should = status_code
 
     # status code
 
@@ -287,14 +326,14 @@ class HTTP:
         """
         Fails if the response does not have a header named `header_name`
         """
-        assert header_name in self.response.headers,\
+        assert header_name in self.response.headers, \
                'Response did not have "%s" header, but should have.' % header_name
 
     def response_should_not_have_header(self, header_name):
         """
         Fails if the response does has a header named `header_name`
         """
-        assert not header_name in self.response.headers,\
+        assert not header_name in self.response.headers, \
                'Response did have "%s" header, but should not have.' % header_name
 
     def get_response_header(self, header_name):
@@ -312,7 +351,7 @@ class HTTP:
         """
         self.response_should_have_header(header_name)
         actual = self.response.headers[header_name]
-        assert actual == expected,\
+        assert actual == expected, \
                'Response header "%s" should have been "%s" but was "%s".' % (
                     header_name, expected, actual)
 
@@ -323,7 +362,7 @@ class HTTP:
         """
         self.response_should_have_header(header_name)
         actual = self.response.headers[header_name]
-        assert actual != not_expected,\
+        assert actual != not_expected, \
                'Response header "%s" was "%s" but should not have been.' % (
                     header_name, actual)
 
@@ -347,7 +386,7 @@ class HTTP:
         `header_value` is the key of the header, e.g. `RobotFramework HttpLibrary (Mozilla/4.0)`
         """
         logger.info('Set request header "%s" to "%s"' % (header_name, header_value))
-        self._request_headers[header_name] = header_value
+        self.context.request_headers[header_name] = header_value
 
     def set_basic_auth(self, username, password):
         """
@@ -375,7 +414,7 @@ class HTTP:
         | Response Should Succeed  |                                     |
         """
         logger.info('Request body set to "%s".' % body)
-        self._request_body = body.encode("utf-8")
+        self.context.request_body = body.encode("utf-8")
 
     def get_response_body(self):
         """
@@ -397,7 +436,9 @@ class HTTP:
         | Response Body Should Contain | version="1.0"    |
         | Response Body Should Contain | encoding="UTF-8" |
         """
-        assert should_contain in self.response.body,\
+        logger.debug('Testing whether "%s" contains "%s".' % (self.response.body, should_contain))
+
+        assert should_contain in self.response.body, \
                '"%s" should have contained "%s", but did not.' % (self.response.body, should_contain)
 
     def log_response_body(self, log_level='INFO'):
@@ -518,4 +559,4 @@ class HTTP:
 
         This is meant for debugging response body's with complex media types.
         """
-        self._response.showbrowser()
+        self.context.response.showbrowser()
